# webserver subprocess

import logging
import os
import signal
from typing import Dict, List, Optional, Union
from uuid import UUID, uuid4

import orjson
import requests
from fastapi import (
    BackgroundTasks,
    Body,
    Depends,
    FastAPI,
    Header,
    HTTPException,
    Request,
    Response,
)
from pydantic import BaseModel, Field
from starlette.middleware import Middleware

from . import settings
from .content_size_limit import ContentSizeLimitMiddleware
from .event_queue import EventQueue
from .signature import Signature
from .validate import validate_signature

logger = logging.getLogger(__name__)

debug = logger.debug
info = logger.info
warning = logger.warning
error = logger.error
critical = logger.critical


app = FastAPI(
    middleware=[
        Middleware(
            ContentSizeLimitMiddleware,
            max_content_size=settings.MAX_CONTENT_SIZE,
        )
    ],
    dependencies=[Depends(validate_signature)],
)

# models


class EventQueueFactory:
    events = EventQueue()

    @classmethod
    async def get_events(self):
        return self.events


async def get_event_list():
    return await EventQueueFactory.get_events()


def orjson_dumps(v, *, default):
    return orjson.dumps(v, default=default).decode()


class VBaseModel(BaseModel):
    class Config:
        json_loads = orjson.loads
        json_dumps = orjson_dumps

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)
        ret = orjson.loads(orjson_dumps(ret, default=None))
        return ret


class Relay(VBaseModel):
    url: Union[str, None] = Field(..., description="relay target url")
    header: Union[str, None] = Field(
        ..., description="header label for api_key"
    )
    key: Union[str, None] = Field(
        ..., description="relay target api authentication key"
    )


class RelayResponse(VBaseModel):
    result: Relay


class Event(VBaseModel):
    id: UUID = Field(..., description="event id")
    path: str = Field(..., descripton="endpoint path")
    method: str = Field(..., description="HTTP method")
    headers: Dict = Field(..., description="HTTP request headers")
    body: Dict = Field(..., description="body data")
    relay: Optional[Dict] = Field(
        None, description="request return value for forwarded event"
    )


class EventResponse(VBaseModel):
    result: Union[str, Dict, None]


class EventsResponse(VBaseModel):
    result: List[Event]


class MessageResponse(VBaseModel):
    result: Union[str, None]


class BoolResponse(VBaseModel):
    result: bool


class EnableRequest(VBaseModel):
    enable: bool


# helper functions


def header_key(key):
    return "-".join([k.capitalize() for k in key.split("-")])


def obscure_key(key):
    if key is not None:
        key = key[:3] + "...." + key[-3:]
    return key


@app.on_event("startup")
async def startup_event():
    debug(f"{__name__} startup")
    app.state.signature = Signature()
    if not hasattr(app.state, "config"):
        app.state.config = {}
    if not hasattr(app.state, "tunnel_url"):
        app.state.tunnel_url = None
    debug(f"{app.state.config=}")
    debug(f"{app.state.tunnel_url=}")


@app.on_event("shutdown")
async def shutdown_event():
    debug(f"{__name__} shutdown")


@app.get("/hello", response_model=MessageResponse)
async def get_hello():
    return MessageResponse(result="Hello, World!")


# @app.get("/tunnel", response_model=MessageResponse)
@app.get("/tunnel", response_model=MessageResponse)
async def get_tunnel(request: Request):
    tunnel_url = request.app.state.tunnel_url
    return MessageResponse(result=tunnel_url)


@app.post("/buffer", response_model=BoolResponse)
async def post_buffer(
    request: EnableRequest, events: EventQueue = Depends(get_event_list)
):
    events.buffer_enabled = request.enable
    return BoolResponse(result=events.buffer_enabled)


@app.get("/buffer", response_model=BoolResponse)
async def get_buffer(events: EventQueue = Depends(get_event_list)):
    return BoolResponse(result=events.buffer_enabled)


@app.post("/relay", response_model=RelayResponse)
async def post_relay(
    request: Relay, events: EventQueue = Depends(get_event_list)
):
    events.relay_url = request.url
    events.relay_header = request.header
    events.relay_key = request.key

    return RelayResponse(
        result=Relay(
            url=events.relay_url,
            header=events.relay_header,
            key=obscure_key(events.relay_key),
        )
    )


@app.get("/relay", response_model=RelayResponse)
async def get_relay(events: EventQueue = Depends(get_event_list)):
    return RelayResponse(
        result=Relay(
            url=events.relay_url,
            header=events.relay_header,
            key=obscure_key(events.relay_key),
        )
    )


@app.post("/contract/event", response_model=EventResponse)
async def post_contract_event(
    event: Dict, request: Request, events: EventQueue = Depends(get_event_list)
):
    baselen = len(str(request.base_url))

    event = Event(
        id=str(uuid4()),
        path=str(request.url)[baselen:],
        method=request.method,
        headers=dict(request.headers),
        body=event,
    )
    event_id = events.append(event)
    return EventResponse(result=str(event_id))


@app.get("/events", response_model=EventsResponse)
async def get_events(events: EventQueue = Depends(get_event_list)):
    return EventsResponse(result=events.list())


@app.get("/clear", response_model=MessageResponse)
async def get_clear(events: EventQueue = Depends(get_event_list)):
    return MessageResponse(result=events.clear())


@app.get("/event/{event_id}", response_model=EventResponse)
async def get_event(
    response: Response,
    event_id: UUID,
    events: EventQueue = Depends(get_event_list),
):
    event = events.lookup(event_id, delete=False)
    if event is None:
        response.status_code = requests.codes.NOT_FOUND
    return EventResponse(result=event)


@app.delete("/event/{event_id}", response_model=EventResponse)
async def delete_event(
    response: Response,
    event_id: UUID,
    events: EventQueue = Depends(get_event_list),
):
    event = events.lookup(event_id, delete=True)
    if event is None:
        response.status_code = requests.codes.NOT_FOUND
    return EventResponse(result=event)


def reaper():
    os.killpg(os.getpgid(os.getpid()), signal.SIGTERM)


@app.get("/shutdown", response_model=MessageResponse)
async def get_shutdown(background_tasks: BackgroundTasks):
    background_tasks.add_task(reaper)
    return MessageResponse(result="shutdown pending")



# defaults module

STREAMS_URL = "https://streams-api.aws-prod-streams-master-1.moralis.io"
SERVER_ADDR = "127.0.0.1"
SERVER_PORT = 8080
QSIZE = 1024
ROW_LIMIT = 100
PAGE_LIMIT = 10000
REGION_CHOICES = ["us-east-1", "us-west-2", "eu-central-1", "ap-southeast-1"]
REGION = REGION_CHOICES[0]
ACTIVE = "active"
PAUSED = "paused"
ERROR = "error"
STATUS_CHOICES = [ACTIVE, PAUSED, ERROR]



from pathlib import Path

from starlette.config import Config
from starlette.datastructures import Secret

from . import defaults

# load variables from .env if it exists
config = Config(".env")

DEBUG = config("WEBHOOK_DEBUG", cast=bool, default=False)
API_KEY = config("WEBHOOK_API_KEY", cast=Secret)
BUFFER_ENABLE = config("WEBHOOK_BUFFER_ENABLE", cast=bool, default=True)
RELAY_URL = config("WEBHOOK_RELAY_URL", cast=str, default=None)
RELAY_HEADER = config("WEBHOOK_RELAY_HEADER", default="X-API-Key")
RELAY_ID_HEADER = config("WEBHOOK_RELAY_ID_HEADER", default="X-Relay-ID")
RELAY_KEY = config("WEBHOOK_RELAY_KEY", cast=Secret)
ADDR = config("WEBHOOK_ADDR", cast=str, default="0.0.0.0")
PORT = config("WEBHOOK_PORT", cast=int, default=8080)
TUNNEL = config("WEBHOOK_TUNNEL", cast=bool, default=True)
NGROK_AUTHTOKEN = config("NGROK_AUTHTOKEN", cast=Secret)
LOG_LEVEL = config("WEBHOOK_LOG_LEVEL", cast=str, default="WARNING")
MAX_CONTENT_SIZE = config(
    "WEBHOOK_MAX_CONTENT_SIZE", cast=int, default=10_000_000
)



# server tests

import json
from uuid import UUID, uuid4

import pytest
from fastapi.testclient import TestClient

from moralis_streams_client import settings
from moralis_streams_client.app import app, obscure_key
from moralis_streams_client.signature import Signature


@pytest.fixture
def client(webhook_process):
    with TestClient(app) as client:
        yield client


@pytest.fixture
def signature():
    key = str(settings.API_KEY)
    return Signature(key)


@pytest.fixture
def _do_request(client, signature):
    def _do_request(method, path, headers={}, data={}, expect_status=200):
        headers.update(signature.headers(data))
        response = client.request(method, path, headers=headers, json=data)
        assert response.status_code == expect_status
        response_dict = response.json()
        if response.ok:
            assert isinstance(response_dict, dict)
            assert set(response_dict.keys()) == set(["result"])
            result = response_dict["result"]
        else:
            result = response_dict
        return result

    assert _do_request("GET", "clear", {}, {}) == "cleared"

    return _do_request


@pytest.fixture
def get(_do_request):
    def _get(path, **kwargs):
        return _do_request("GET", path, **kwargs)

    return _get


@pytest.fixture
def post(_do_request, get):
    def _post(path, **kwargs):
        return _do_request("POST", path, **kwargs)

    return _post


@pytest.fixture
def delete(_do_request, get):
    def _delete(path, **kwargs):
        return _do_request("DELETE", path, **kwargs)

    return _delete


def test_server_hello(get):
    result = get("hello")
    assert result == "Hello, World!"


def test_server_tunnel(get):
    result = get("tunnel")
    assert result is None


def test_server_buffer(post, get):
    assert get("buffer") is True
    assert post("buffer", data=dict(enable=False)) is False
    assert get("buffer") is False
    assert post("buffer", data=dict(enable=True)) is True
    assert get("buffer") is True


def _obscured(relay):
    ret = {}
    for k, v in relay.items():
        if k == "key":
            if v is not None:
                v = obscure_key(v)
        ret[k] = v
    return ret


def test_server_relay(get, post):
    default_relay = dict(
        url=settings.RELAY_URL,
        header=settings.RELAY_HEADER,
        key=str(settings.RELAY_KEY),
    )

    test_relay = dict(
        url="http://howdy/howdy/howdy",
        header="x-crazy-header",
        key="when-you-can-snatch-the-pebble-from-my-hand",
    )

    null_relay = dict(url=None, header=None, key=None)

    response = get("relay")
    assert response == _obscured(default_relay)

    response = post("relay", data=test_relay)
    assert response == _obscured(test_relay)

    response = post("relay", data=null_relay)
    assert response == null_relay

    response = post("relay", data=default_relay)
    assert response == _obscured(default_relay)


@pytest.fixture
def testevent():
    return dict(
        id=str(uuid4()),
        kniggits=["robin", "bedeveire", "lancelot"],
        swallows=["african", "european"],
        data=[dict(id=str(uuid4()), sketch="confuse-a-cat")],
        more_data=dict(foo=dict(id=str(uuid4())), id=str(uuid4())),
        grip_method="by the husk",
    )


def test_server_contract_event(get, post, testevent):
    assert get("events") == []
    eid = post("contract/event", data=testevent)
    assert UUID(eid)
    events = get("events")
    assert len(events) == 1
    assert events[0]["id"] == eid


def test_server_events(get, post, testevent):
    assert get("events") == []
    eid1 = post("contract/event", data=testevent)
    assert UUID(eid1)
    eid2 = post("contract/event", data=testevent)
    assert UUID(eid2)
    eid3 = post("contract/event", data=testevent)
    assert UUID(eid3)
    events = get("events")
    assert isinstance(events, list)
    assert len(events) == 3
    for event in events:
        assert isinstance(event, dict)
        assert UUID(event["id"])


def test_server_clear(get):
    result = get("clear")
    assert result == "cleared"


def test_server_get_event(get, post, testevent):
    assert get("events") == []
    event_id = post("contract/event", data=testevent)
    assert UUID(event_id)
    event = get(f"event/{event_id}")
    assert event["body"] == testevent
    not_found = get("events/blarg", expect_status=404)
    assert isinstance(not_found, dict)
    assert "result" not in not_found


def test_server_delete_event(get, post, delete, testevent):
    assert get("events") == []
    event_id = post("contract/event", data=testevent)
    assert UUID(event_id)
    events = get("events")
    assert len(events) == 1
    assert event_id == events[0]["id"]
    deleted_event = delete(f"event/{event_id}")
    assert deleted_event["body"] == testevent
    events = get("events")
    assert events == []


def test_server_send_relay(get, post, testevent, webhook_process_class):
    with webhook_process_class(
        port=8081, tunnel=False, logfile="target.log"
    ) as target:
        assert target.server_running()
        assert target.clear() == "cleared"
        assert target.events() == []
        relay_config = dict(
            url="http://localhost:8081/contract/event",
            header="X-API-Key",
            key="test_server_send_relay",
        )
        ret = post("relay", data=relay_config)
        assert ret["url"] == relay_config["url"]
        event_id = post("contract/event", data=testevent)

        sent_events = get("events")
        assert len(sent_events) == 1
        assert sent_events[0]["id"] == event_id
        assert sent_events[0]["relay"]

        events = target.events()
        assert len(events) == 1
        assert events[0]["headers"]["x-relay-id"] == event_id
        assert events[0]["headers"]["x-api-key"] == relay_config["key"]
        assert events[0]["body"] == testevent



import collections

import requests
from requests.structures import CaseInsensitiveDict

from . import defaults, settings


class EventQueue:
    def __init__(self):
        self.events = collections.deque([])
        self.buffer_enabled = settings.BUFFER_ENABLE
        self.relay_url = settings.RELAY_URL
        self.relay_header = settings.RELAY_HEADER
        self.relay_key = str(settings.RELAY_KEY)
        self.relay_id_header = settings.RELAY_ID_HEADER

    def append(self, event):
        if self.relay_url:
            _response = self.forward(event)
            event.relay = dict(
                url=_response.url,
                status_code=_response.status_code,
                text=_response.text,
                headers=dict(_response.headers),
            )
        else:
            event.relay = None
        if self.buffer_enabled:
            self.events.append(event)
        return event.id

    def forward(self, event):
        headers = event.headers
        if self.relay_header and self.relay_key:
            headers[self.relay_header] = self.relay_key
        headers[self.relay_id_header] = str(event.id)
        return requests.post(self.relay_url, headers=headers, json=event.body)

    def list(self):
        # create a list from the event_queue, preserving order
        ret = []
        tq = self.events.copy()
        try:
            while True:
                ret.append(tq.popleft())
        except IndexError:
            pass
        return ret

    def clear(self):
        self.events.clear()
        return "cleared"

    def lookup(self, event_id, delete):
        tq = self.events.copy()
        if delete:
            self.events.clear()
        found = None
        while True:
            try:
                event = tq.popleft()
                if event.id == event_id:
                    found = event
                    if delete is False:
                        break
                elif delete is True:
                    self.events.append(event)
            except IndexError:
                break

        return found



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Extra, Field


class VBaseModel(BaseModel):
    class Config:
        extra = Extra.forbid
        json_encoders = {"UUIDModel": lambda b: str(b.__root__)}


class UUIDModel(VBaseModel):
    __root__: UUID = Field(
        ...,
        description="Stringified UUIDv4.\nSee [RFC 4112](https://tools.ietf.org/html/rfc4122)",
    )


class Block(VBaseModel):

    number: str
    hash: str
    timestamp: str


class Log(VBaseModel):

    logIndex: str
    transactionHash: str
    address: str
    data: str
    topic0: str
    topic1: str
    topic2: str
    topic3: str


class Transaction(VBaseModel):

    hash: str
    gas: str
    gasPrice: str
    nonce: str
    input: str
    transactionIndex: str
    fromAddress: str
    toAddress: str
    value: str
    type: str
    v: str
    r: str
    s: str
    receiptCumulativeGasUsed: str
    receiptGasUsed: str
    receiptContractAddress: str
    receiptRoot: str
    receiptStatus: str


class InternalTransaction(VBaseModel):

    from_: str = Field(..., alias="from")
    to: str
    value: str
    transactionHash: str
    gas: str


class AbiInput(VBaseModel):

    name: str
    type: str
    indexed: Optional[bool] = None
    components: Optional[List[AbiInput]] = None
    internalType: Optional[str] = None


class AbiOutput(VBaseModel):

    name: str
    type: str
    components: Optional[List[AbiOutput]] = None
    internalType: Optional[str] = None


class AbiItem(VBaseModel):

    anonymous: Optional[bool] = None
    constant: Optional[bool] = None
    inputs: Optional[List[AbiInput]] = None
    name: Optional[str] = None
    outputs: Optional[List[AbiOutput]] = None
    payable: Optional[bool] = None
    stateMutability: Optional[str] = None
    type: str
    gas: Optional[float] = None


class HistoryModel(VBaseModel):

    id: UUIDModel
    date: datetime
    payload: Optional[IWebhookUnParsed] = None
    tinyPayload: ITinyPayload
    errorMessage: str
    webhookUrl: str
    streamId: str
    tag: str


class SettingsRegion(Enum):
    us_east_1 = "us-east-1"
    us_west_2 = "us-west-2"
    eu_central_1 = "eu-central-1"
    ap_southeast_1 = "ap-southeast-1"


class StreamsStatus(Enum):
    active = "active"
    paused = "paused"
    error = "error"
    terminated = "terminated"


class StreamsFilter(VBaseModel):
    pass


class AdvancedOptions(VBaseModel):

    topic0: str
    filter: Optional[StreamsFilter] = None
    includeNativeTxs: Optional[bool] = None


class StreamsModel(VBaseModel):

    webhookUrl: str = Field(
        ...,
        description="Webhook URL where moralis will send the POST request.",
    )
    description: str = Field(..., description="A description for this stream")
    tag: str = Field(
        ...,
        description="A user-provided tag that will be send along the webhook, the user can use this tag to identify the specific stream if multiple streams are present",
    )
    topic0: Optional[List[str]] = Field(
        None,
        description="An Array of topic0's in hex, required if the type : log",
    )
    allAddresses: Optional[bool] = Field(
        None,
        description="Include events for all addresses (only applied when abi and topic0 is provided)",
    )
    includeNativeTxs: Optional[bool] = Field(
        None,
        description="Include or not native transactions defaults to false (only applied when type:contract)",
    )
    includeContractLogs: Optional[bool] = Field(
        None,
        description="Include or not logs of contract interactions defaults to false",
    )
    includeInternalTxs: Optional[bool] = Field(
        None,
        description="Include or not include internal transactions defaults to false",
    )
    abi: Optional[List[AbiItem]] = None
    advancedOptions: Optional[List[AdvancedOptions]] = None
    chainIds: List[str] = Field(
        ...,
        description='The ids of the chains for this stream in hex Ex: ["0x1","0x38"]',
    )
    id: UUIDModel = Field(..., description="The unique uuid of the stream")
    status: StreamsStatus = Field(..., description="The status of the stream.")
    statusMessage: str = Field(
        ..., description="Description of current status of stream."
    )


class Addresses(VBaseModel):

    address: str = Field(..., description="Address")


AbiInput.update_forward_refs()
AbiOutput.update_forward_refs()


class IWebhookUnParsed(VBaseModel):

    block: Block
    chainId: str
    logs: List[Log]
    txs: List[Transaction]
    txsInternal: List[InternalTransaction]
    abi: List[AbiItem]
    retries: float
    confirmed: bool
    tag: str
    streamId: str


class ITinyPayload(VBaseModel):

    chainId: str
    confirmed: bool
    block: str
    records: float
    retries: float



# api tests

import logging
import os
import time
from decimal import Decimal
from pathlib import Path

import MoralisSDK.api
import pytest
from backoff import expo, on_exception
from box import Box, BoxList
from eth_account import Account
from eth_utils import (
    event_abi_to_log_topic,
    from_wei,
    is_same_address,
    to_checksum_address,
    to_hex,
    to_wei,
)
from ratelimit import RateLimitException, limits

import moralis_streams_client
from moralis_streams_client import models

# TODO: generate history and multiple streams to ensure enough response data

debug = logging.debug
info = logging.info
logging.getLogger("moralis_streams_client.api").setLevel("WARNING")


ETHERSCAN_LIMIT_COUNT = 5
ETHERSCAN_LIMIT_SECONDS = 1

CONFIRM_COUNT = 12

EVENT_NAME = "PrintMintPending"

CALLBACK_TIMEOUT = 300

CALLBACK_TEST_KEYS = [
    "abi",
    "block",
    "chainId",
    "confirmed",
    "erc20Approvals",
    "erc20Transfers",
    "logs",
    "nftApprovals",
    "nftTransfers",
    "retries",
    "txs",
    "txsInternal",
    "streamId",
    "tag",
]


@pytest.fixture
def event_name():
    return EVENT_NAME


@pytest.fixture
def background_contract_address(config):
    address = config("TEST_BACKGROUND_CONTRACT")
    return to_checksum_address(address)


# rate-limit etherscan api calls
@on_exception(expo, RateLimitException, max_tries=16)
@limits(calls=ETHERSCAN_LIMIT_COUNT, period=ETHERSCAN_LIMIT_SECONDS)
def limit(func, *args, **kwargs):
    return func(*args, **kwargs)


@pytest.fixture
def get_contract_tokens(moralis, network):
    def _get_contract_tokens(contract_address):
        owners = Box(moralis.get_nft_owners(contract_address, chain=network))
        assert "result" in owners
        return owners.result

    return _get_contract_tokens


def test_api_transactions(ape, get_contract_tokens, system_address, moralis):
    assert ape.explorer
    txns = list(limit(ape.explorer.get_account_transactions, system_address))
    assert txns
    contracts = set()
    for txn in txns:
        info(f"{txn.transaction}")
        to_addr = txn.transaction.dict()["to"]
        if to_addr == "":
            assert txn.contract_address
            contracts.add(to_checksum_address(txn.contract_address))

    info(f"txn_count={len(txns)}")
    info(f"contracts={len(contracts)}")
    for addr in contracts:
        contract = ape.Contract(addr)
        info(f"{contract.symbol()} {contract}")

        url = ape.explorer.get_address_url(addr)
        info(url)

        token_ids = Box(moralis.get_token_id(addr, chain="goerli"))
        info(f"  tokens={int(token_ids.total)}")

        tokens = get_contract_tokens(addr)
        assert token_ids.total == len(tokens)
        for token in tokens:
            info(f"  {int(token.token_id)}: {token.name} {token.owner_of}")


@pytest.fixture
def ethersieve_contract(ape, ethersieve_contract_address):
    return ape.Contract(ethersieve_contract_address)


@pytest.fixture
def background_contract(ape, background_contract_address):
    return ape.Contract(background_contract_address)


@pytest.mark.uses_gas
def test_api_create_stream(
    streams,
    dump,
    chain_id,
    ethersieve_contract,
    background_contract,
    event_name,
    event_keys,
    get_contract_tokens,
    system_address,
    system_key,
    user_address,
    user_key,
    ape,
    webhook,
    webhook_tunnel_url,
):
    webhook.clear()

    streams_begin = streams.get_streams()

    event_abi = ethersieve_contract.contract_type.events[event_name].dict()
    event_topic = to_hex(event_abi_to_log_topic(event_abi))
    contract_abi = ethersieve_contract.contract_type.dict()["abi"]

    stream = Box(
        streams.create_stream(
            webhook_url=webhook_tunnel_url + "/contract/event",
            description="moralis_streams_client testing stream",
            tag="msc_test",
            topic0=[event_topic],
            all_addresses=False,
            include_native_txs=True,
            include_contract_logs=True,
            include_internal_txs=True,
            abi=contract_abi,
            advanced_options=None,
            chain_ids=[chain_id],
        )
    )
    dump(f"{stream.id=}")
    dump(f"{stream.tag=}")
    dump(f"{stream.status=}")
    dump(f"{stream.statusMessage=}")
    address_added = Box(
        streams.add_address_to_stream(stream.id, [ethersieve_contract.address])
    )
    assert isinstance(address_added, dict)
    dump(address_added)
    assert "streamId" in address_added
    assert "address" in address_added
    assert isinstance(address_added.streamId, str)
    assert isinstance(address_added.address, list)
    addresses = address_added.address
    for address in addresses:
        assert isinstance(address, str)
        assert is_same_address(address, ethersieve_contract.address)

    background_ids = []
    for token in get_contract_tokens(background_contract.address):
        if is_same_address(token.owner_of, user_address):
            background_ids.append(token.token_id)
    dump(f"user background tokens: {background_ids}")

    ethersieve_ids = []
    for token in get_contract_tokens(ethersieve_contract.address):
        if is_same_address(token.owner_of, user_address):
            ethersieve_ids.append(token.token_id)
    dump(f"user ethersieve tokens: {ethersieve_ids}")

    # send a printMint transaction from user to ethersieve

    PRIORITY_FEE_GWEI = 3

    user = Account().from_key(user_key)
    max_priority_fee = ape.provider.priority_fee + to_wei(
        PRIORITY_FEE_GWEI, "gwei"
    )

    info(f"max_priority_fee={from_wei(max_priority_fee, 'gwei')} gwei")

    args = (0, 0, 1, 2, 3)
    kwargs = dict(
        value=to_wei(Decimal(".01"), "ether"),
        sender=user.address,
        nonce=ape.provider.get_nonce(user.address),
        max_priority_fee=max_priority_fee,
        required_confirmations=0,
    )
    txn = ethersieve_contract.printMint.as_transaction(*args, **kwargs)
    txn.signature = user.sign_transaction(txn.dict())
    estimated_gas = ape.provider.estimate_gas_cost(txn)
    info(f"{estimated_gas=}")
    receipt = ape.provider.send_transaction(txn)

    dump(dict(txn_hash=receipt.txn_hash, status=str(receipt.status)))

    CONFIRM_COUNT = 2

    confirmed_height = receipt.block_number + CONFIRM_COUNT
    last_height = receipt.block_number

    info(f"Waiting for {CONFIRM_COUNT} confirmations...")
    count = 1
    while last_height < confirmed_height:
        height = ape.provider.chain_manager.blocks.height
        if last_height != height:
            info(f"block[{count} of {CONFIRM_COUNT}]: {height}")
            last_height = height
            count += 1

    info(f"Waiting up to {CALLBACK_TIMEOUT} seconds for 3 callback events...")
    events = {}

    timeout = time.time() + CALLBACK_TIMEOUT
    while len(events) < 3:
        assert time.time() < timeout, "timeout waiting for 3 callback events"
        all_events = webhook.events()
        for event in all_events:
            event_id = event["id"]
            if event_id not in events:
                events[event_id] = event
                info(f"{event_id=}")
        time.sleep(1)

    events = webhook.events()
    assert isinstance(events, list)
    for event in events:
        assert isinstance(event, dict)
        assert set(event.keys()) == set(event_keys)

    assert set(events[0]["body"].keys()) == set(CALLBACK_TEST_KEYS)
    unconfirmed = events[1]["body"]
    assert unconfirmed["confirmed"] is False
    confirmed = events[2]["body"]
    assert confirmed["confirmed"] is True
    for e in [unconfirmed, confirmed]:
        assert len(e["txs"]) == 1
        assert is_same_address(e["txs"][0]["fromAddress"], user_address)
        assert is_same_address(
            e["txs"][0]["toAddress"], ethersieve_contract.address
        )

    result = Box(streams.delete_stream(stream.id))
    assert result.id == stream.id

    streams_end = streams.get_streams()

    assert len(streams_begin) == len(streams_end)


def test_api_get_streams_default(streams):
    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    for i, stream in enumerate(all_streams):
        assert isinstance(stream, dict)
        info(f"{i}: {stream['id']}")


def test_api_get_streams_row_limit(streams, monkeypatch):
    # set low row limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 3)
    streams = streams.get_streams()
    assert isinstance(streams, list)
    for i, stream in enumerate(streams):
        assert isinstance(stream, dict)
        info(f"{i}: {stream['id']}")


def test_api_get_streams_row_page_limit(streams, monkeypatch):
    # set low row and page limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 2)
    monkeypatch.setattr(streams, "page_limit", 2)
    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    for i, stream in enumerate(all_streams):
        assert isinstance(stream, dict)
        info(f"{i}: {stream['id']}")


def test_api_get_history_default(streams):
    history_events = streams.get_history()
    assert isinstance(history_events, list)
    for i, history_event in enumerate(history_events):
        assert isinstance(history_event, dict)
        info(f"{i}: {history_event['id']}")


def test_api_get_history_row_limit(streams, monkeypatch):
    # set low row limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 3)
    history_events = streams.get_history()
    assert isinstance(history_events, list)
    for i, history_event in enumerate(history_events):
        assert isinstance(history_event, dict)
        info(f"{i}: {history_event['id']}")


def test_api_get_history_row_page_limit(streams, monkeypatch):
    # set low row and page limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 3)
    monkeypatch.setattr(streams, "page_limit", 2)
    history_events = streams.get_history()
    assert isinstance(history_events, list)
    for i, history_event in enumerate(history_events):
        assert isinstance(history_event, dict)
        info(f"{i}: {history_event['id']}")


def test_api_delete_all_streams(streams, dump):
    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    for stream in all_streams:
        assert isinstance(stream, dict)
        stream_id = stream["id"]
        dump(f"deleting stream {stream_id}")
        result = streams.delete_stream(stream_id)
        assert isinstance(result, dict)
        assert result["id"] == stream_id

    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    assert len(all_streams) == 0



# webhook server process

import json
import logging
import os
import sys
from pathlib import Path
from subprocess import CalledProcessError, check_output

import click
import requests
from eth_hash.auto import keccak
from eth_utils import to_hex

from . import settings
from .exception_handler import ExceptionHandler
from .signature import Signature
from .webhook import Webhook

logger = logging.getLogger(__name__)
debug = logger.debug


def output(
    data,
):
    click.echo(json.dumps(data, indent=2))


def fail(error):
    click.echo(f"Error: {error}", err=True)
    sys.exit(-1)


def run(cmd):
    return check_output(cmd, shell=True).decode()


@click.group
@click.option(
    "-a",
    "--addr",
    type=str,
    envvar="WEBHOOK_SERVER_ADDR",
    default="0.0.0.0",
    show_default=True,
    show_envvar=True,
    help="server listen IP addr",
)
@click.option(
    "-p",
    "--port",
    type=int,
    envvar="WEBHOOK_SERVER_PORT",
    show_default=True,
    default=8080,
    show_envvar=True,
    help="server listen port",
)
@click.option(
    "-t/-T",
    "--tunnel/--no-tunnel",
    is_flag=True,
    default=True,
    help="enable / disable ngrok tunnel",
)
@click.option(
    "-r",
    "--relay-url",
    type=str,
    envvar="WEBHOOK_RELAY_URL",
    show_envvar=True,
    help="enable relay to URL",
)
@click.option(
    "-k",
    "--relay-key",
    type=str,
    envvar="WEBHOOK_RELAY_KEY",
    show_envvar=True,
    help="relay API key",
)
@click.option(
    "-h",
    "--relay-header",
    type=str,
    envvar="WEBHOOK_RELAY_HEADER",
    show_envvar=True,
    help="header name for relay API key",
)
@click.option(
    "-b/-B",
    "--enable-buffer/--disable-buffer",
    is_flag=True,
    default=True,
    help="enable/disable event buffer",
)
@click.option(
    "-m",
    "--moralis-api-key",
    type=str,
    envvar="MORALIS_API_KEY",
    show_envvar=True,
    help="Moralis Streams API key",
)
@click.option("-d", "--debug", is_flag=True)
@click.option(
    "-l",
    "--log-level",
    type=str,
    default="WARNING",
    envvar="WEBHOOK_LOG_LEVEL",
    help="logging level",
)
@click.pass_context
def webhook(
    ctx,
    addr,
    port,
    debug,
    tunnel,
    relay_url,
    relay_key,
    relay_header,
    enable_buffer,
    moralis_api_key,
    log_level,
):
    """webhook endpoint server commands"""
    if debug:
        log_level = "DEBUG"
    logging.basicConfig(level=log_level)
    logger.setLevel(log_level)

    if ctx.obj is None:
        ctx.obj = dict(ehandler=ExceptionHandler(debug))
    kwargs = {}
    kwargs["addr"] = addr
    kwargs["port"] = port
    kwargs["debug"] = debug
    kwargs["tunnel"] = tunnel
    kwargs["relay_url"] = relay_url
    kwargs["relay_key"] = relay_key
    kwargs["relay_header"] = relay_header
    kwargs["enable_buffer"] = enable_buffer
    kwargs["moralis_api_key"] = moralis_api_key
    ctx.obj.update(**kwargs)
    ctx.obj["webhook"] = Webhook(**kwargs)


@webhook.command
@click.pass_context
@click.option("-q", "--quiet", is_flag=True)
def ps(ctx, quiet):
    """test for running server"""
    webhook = ctx.obj["webhook"]
    procs = webhook.processes()
    if bool(procs):
        ret = 0
        if not quiet:
            for proc in procs:
                click.echo(repr(proc))
    else:
        ret = -1
    sys.exit(ret)


@webhook.command
@click.pass_context
def hello(ctx):
    """output the ngrok public url"""
    webhook = ctx.obj["webhook"]
    output(webhook.hello())


@webhook.command
@click.pass_context
def tunnel(ctx):
    """output the ngrok public url"""
    webhook = ctx.obj["webhook"]
    output(webhook.tunnel_url())


@webhook.command
@click.option(
    "-e",
    "--enable",
    "buffer_enabled",
    flag_value=True,
    default=None,
    help="enable event buffer",
)
@click.option(
    "-d",
    "--disable",
    "buffer_enabled",
    flag_value=False,
    default=None,
    help="disable event buffer",
)
@click.pass_context
def buffer(ctx, buffer_enabled):
    """enable or disable the event buffer"""
    if buffer_enabled is None:
        kwargs = {}
    else:
        kwargs = dict(enable=buffer_enabled)
    webhook = ctx.obj["webhook"]
    output(webhook.buffer(**kwargs))


@webhook.command
@click.option(
    "-e",
    "--enable",
    "relay_enabled",
    flag_value=True,
    default=None,
    help="enable event relay",
)
@click.option(
    "-d",
    "--disable",
    "relay_enabled",
    flag_value=False,
    default=None,
    help="disable event relay",
)
@click.pass_context
def relay(ctx, relay_enabled):
    """set the configuration"""
    webhook = ctx.obj["webhook"]
    output(
        webhook.relay(
            url=ctx.obj["relay_url"],
            header=ctx.obj["relay_header"],
            key=ctx.obj["relay_key"],
            enable=relay_enabled,
        )
    )


@webhook.command
@click.option("-m", "--message", type=str, help="create json message data")
@click.argument("input", default="-", type=click.File("r"))
@click.pass_context
def inject(ctx, message, input):
    """inject an event"""
    webhook = ctx.obj["webhook"]
    if message:
        data = dict(message=message)
    else:
        data = json.load(input)
    output(webhook.inject(data))


@webhook.command
@click.argument("event-id", type=str)
@click.pass_context
def event(ctx, event_id):
    """get an event by id"""
    webhook = ctx.obj["webhook"]
    output(webhook.event(event_id))


@webhook.command
@click.argument("event-id", type=str)
@click.pass_context
def delete(ctx, event_id):
    """delete an event by id"""
    webhook = ctx.obj["webhook"]
    output(webhook.delete(event_id))


@webhook.command
@click.pass_context
def clear(ctx):
    """clear the webhook sever event buffer"""
    webhook = ctx.obj["webhook"]
    output(webhook.clear())


@webhook.command
@click.argument("output", default="-", type=click.File("w"))
@click.pass_context
def events(ctx, output):
    """output events received by the webhook server"""
    webhook = ctx.obj["webhook"]
    json.dump(webhook.events(), output, indent=2)
    output.write("\n")


@webhook.command
@click.option(
    "-l",
    "--logfile",
    envvar="TUNNEL_LOGFILE",
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
)
@click.option(
    "-w/-W",
    "--wait/--no-wait",
    is_flag=True,
    default=True,
    help="wait for server startup",
)
@click.pass_context
def start(ctx, logfile, wait):
    """start local webhook server process with optional ngrok tunnel"""
    kwargs = ctx.obj.copy()
    webhook = kwargs.pop("webhook")
    kwargs.pop("ehandler")
    kwargs["logfile"] = logfile
    kwargs["wait"] = wait

    if webhook.server_running():
        fail("already running")
    else:
        click.echo("starting webhook server...", nl=False, err=True)
        proc = webhook.start(logfile, wait)
        if wait:
            while not webhook.server_running():
                if proc.poll() is not None:
                    click.echo("error")
                    fail(f"process {proc.args} returned {proc.returncode}")
                click.echo(".", nl=False, err=True)

        click.echo("started", err=True)


@webhook.command
@click.option(
    "-w/-W",
    "--wait/--no-wait",
    is_flag=True,
    default=True,
    help="wait for server shutdown",
)
@click.pass_context
def stop(ctx, wait):
    """stop the webhook server"""
    webhook = ctx.obj["webhook"]
    if not webhook.server_running():
        fail("not running")
    else:
        click.echo(webhook.shutdown(), err=True, nl=False)
    while webhook.server_running():
        click.echo(".", err=True, nl=False)
    click.echo("stopped", err=True)


@webhook.command
@click.option(
    "--ngrok-token",
    type=str,
    envvar="NGROK_AUTHTOKEN",
    show_envvar=True,
    help="ngrok auth token",
)
@click.option(
    "-w",
    "--workers",
    type=int,
    default=1,
    help="number of worker threads/processes",
)
@click.pass_context
def server(ctx, ngrok_token, workers):
    """run webhook server process"""
    from .server import ServerProcess

    config = ctx.obj.copy()
    config.pop("ehandler")
    config["ngrok_token"] = ngrok_token
    config["workers"] = workers
    ServerProcess(**config).run()



# signature tests

import json
from logging import info
from pathlib import Path
from pprint import pformat

import pytest

from moralis_streams_client import settings
from moralis_streams_client.signature import Signature


@pytest.fixture
def testdata():
    return dict(
        parrot=dict(breed="norweigan blue", status="pining for the fjords")
    )


@pytest.fixture
def testbytes(testdata):
    return json.dumps(testdata).encode()


@pytest.fixture
def testkey():
    return "this_parrot_has_ceased_to_be".encode()


@pytest.fixture
def testheader():
    return "X-Parrot-Test"


@pytest.fixture
def badbytes():
    return "and_now_for_something_completely_different".encode()


def test_signature_defaults(testkey):
    s = Signature(testkey)
    assert s.key == testkey
    assert s.header.startswith("X-")


def test_signature_validation(testkey, testbytes, badbytes):
    s = Signature(testkey)
    sig = s.calculate(testbytes)
    valid = s.validate(sig, testbytes)
    assert valid is True
    invalid_data = s.validate(sig, badbytes)
    assert invalid_data is False
    invalid_key = s.validate(badbytes, testbytes)
    assert invalid_key is False
    valid_again = s.validate(sig, testbytes)
    assert valid_again is True


def test_signature_headers(testkey, testheader, testbytes):
    s = Signature(testkey, testheader)
    sig = s.calculate(testbytes)
    headers = s.headers(testbytes)
    assert headers == {testheader: sig}
    valid = s.validate(headers[testheader], testbytes)
    assert valid is True


def test_signature_str_key(testbytes):
    testkey = "dimsdale"
    s = Signature(testkey)
    sig = s.calculate(testbytes)
    valid = s.validate(sig, testbytes)
    assert valid is True


def test_signature_str_sig(testbytes):
    testkey = "a_lupin"
    s = Signature(testkey)
    sig = s.calculate(testbytes)
    assert isinstance(sig, str)
    assert s.validate(sig, testbytes)
    assert not s.validate("completely_different", testbytes)


def test_signature_sample_event(shared_datadir):
    event_file = shared_datadir / "event.json"
    if not event_file.is_file():
        info("copy a captured moralis streams event to tests/data/event.json")
        return
    sample_event = json.loads(event_file.read_text())
    assert isinstance(sample_event, dict)
    info(pformat(sample_event))

    sample = sample_event["headers"]["x-signature"]
    key = str(settings.API_KEY).encode()
    s = Signature(key)
    body = sample_event["body"]

    data = body
    # data=json.dumps(body,separators=(',',':')).encode()

    local = s.calculate(data)
    headers = s.headers(data)

    info(f"{key=}")
    info(f"{data=}")
    info(f"{sample=}")
    info(f"{local=}")
    info(f"{headers=}")



# streams api client wrapper

import json
import logging
import os

from eth_hash.auto import keccak
from eth_utils import to_hex

from . import settings

logger = logging.getLogger("signature")
debug = logger.debug
error = logger.error
critical = logger.critical


class Signature:
    def __init__(self, key=None, header="X-Signature"):
        key = key or str(settings.API_KEY)
        self.key = self._bytes(key)
        self.header = header

    def _bytes(self, data):
        if isinstance(data, dict):
            data = json.dumps(data, separators=(",", ":")).encode()
        if isinstance(data, str):
            data = data.encode()
        elif not isinstance(data, bytes):
            raise TypeError()
        return data

    def validate(self, signature: bytes, body: bytes) -> bool:
        """calculate the sha3 checksum and validate the request"""
        valid = signature == self.calculate(body)
        return valid

    def calculate(self, body: bytes) -> bytes:
        """calculate the sha3 checksum of body and api_key"""
        body = self._bytes(body)
        debug(f"calculate: {len(body)} bytes {body=}")
        body = self._bytes(body)
        s = keccak.new(body)
        s.update(self.key)
        ret = to_hex(s.digest())
        debug(f"{ret=}")
        return ret

    def headers(self, body: bytes) -> dict:
        return {self.header: self.calculate(body)}



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from uuid import UUID

from pydantic import BaseModel, Field

from .model import VBaseModel



# autogenerated models

from .addressesTypes import (
    AddressesAdd,
    AddressesRemove,
    AddressesResponse,
    AddressResponse,
    DeleteAddressResponse,
)
from .historyTypes import HistoryResponse
from .model import (
    AbiInput,
    AbiItem,
    AbiOutput,
    Addresses,
    AdvancedOptions,
    Block,
    HistoryModel,
    InternalTransaction,
    Log,
    SettingsRegion,
    StreamsFilter,
    StreamsModel,
    StreamsStatus,
    Transaction,
    UUIDModel,
)
from .Partial_streamsTypes import StreamsModelUpdate
from .settingsTypes import SettingsModel
from .statsTypes import StatsModel
from .streamsTypes import (
    StreamsModelCreate,
    StreamsResponse,
    StreamsStatusUpdate,
)

__all__ = [
    "UUIDModel",
    "Block",
    "Log",
    "Transaction",
    "InternalTransaction",
    "AbiInput",
    "AbiOutput",
    "AbiItem",
    "HistoryModel",
    "SettingsRegion",
    "StreamsStatus",
    "StreamsFilter",
    "AdvancedOptions",
    "StreamsModel",
    "Addresses",
    "StreamsResponse",
    "StreamsModelCreate",
    "StreamsStatusUpdate",
    "SettingsModel",
    "StatsModel",
    "HistoryResponse",
    "AddressesResponse",
    "AddressResponse",
    "AddressesAdd",
    "DeleteAddressResponse",
    "AddressesRemove",
    "StreamsModelUpdate",
]



# streams api exceptions


class MoralisStreamsError(Exception):
    pass


class MoralisStreamsErrorReturned(MoralisStreamsError):
    pass


class MoralisStreamsCallFailed(MoralisStreamsError):
    pass


class MoralisStreamsResponseFormatError(MoralisStreamsError):
    pass



# Signature checking middleware

import json
import logging

from fastapi import HTTPException
from requests import codes
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import PlainTextResponse

from .signature import Signature

logger = logging.getLogger(__name__)
debug = logger.debug
error = logger.error
critical = logger.critical


class SignatureValidator:
    signature = None

    def __init__(self, signature):
        self.__class__.signature = signature

    async def set_cached_receive(self, request):
        cached = await request._receive()

        async def receive():
            return cached

        request._receive = receive



#!/usr/bin/env python

"""Tests for `moralis_streams_client` CLI"""

import json
import shlex
from traceback import print_exception

import pytest
from click.testing import CliRunner

import moralis_streams_client
from moralis_streams_client import __version__
from moralis_streams_client.cli import cli


def test_version():
    """Test reading version and module name"""
    assert moralis_streams_client.__name__ == "moralis_streams_client"
    assert __version__
    assert isinstance(__version__, str)


@pytest.fixture
def run():
    runner = CliRunner()

    def _run(*args, **kwargs):
        if len(args) == 1 and isinstance(args[0], list):
            cmd = args[0]
        elif len(args) > 0:
            cmd = shlex.split(" ".join(list(args)))
        else:
            raise RuntimeError(f"args unexpected: {repr(args)}")
        expect_json = kwargs.pop("expect_json", False)
        expect_exit_code = kwargs.pop("expect_exit_code", 0)
        expect_exception = kwargs.pop("expect_exception", None)
        result = runner.invoke(cli, cmd, **kwargs)
        if result.exception:
            if expect_exception and isinstance(
                result.exception, expect_exception
            ):
                pass
            else:
                raise result.exception from result.exception
        else:
            assert result.exit_code == expect_exit_code, result.output
        if expect_json:
            result = json.loads(result.stdout)
        return result

    return _run


def test_cli_none(run):
    result = run([])
    assert "Usage:" in result.output


def test_cli_help(run):
    result = run(["--help"])
    assert "Show this message and exit." in result.output



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from pydantic import BaseModel, Extra, Field

from .model import VBaseModel


class StatsModel(VBaseModel):

    totalWebhooksDelivered: float = Field(
        ...,
        description="The total amount of webhooks delivered across all streams",
    )
    totalWebhooksFailed: float = Field(
        ...,
        description="The total amount of failed webhooks across all streams",
    )
    totalLogsProcessed: float = Field(
        ...,
        description="The total amount of logs processed across all streams, this includes failed webhooks",
    )
    totalTxsProcessed: float = Field(
        ...,
        description="The total amount of txs processed across all streams, this includes failed webhooks",
    )
    totalTxsInternalProcessed: float = Field(
        ...,
        description="The total amount of internal txs processed across all streams, this includes failed webhooks",
    )



"""Client module for the Moralis Streams API, including a CLI and a webhook utility for buffering and forwarding endpoint callbacks."""

from .api import MoralisStreamsApi
from .defaults import ACTIVE, ERROR, PAUSED, REGION_CHOICES, STATUS_CHOICES
from .exceptions import (
    MoralisStreamsCallFailed,
    MoralisStreamsError,
    MoralisStreamsErrorReturned,
    MoralisStreamsResponseFormatError,
)
from .version import __author__, __email__, __timestamp__, __version__

__all__ = [
    __version__,
    __timestamp__,
    __author__,
    __email__,
    "ACTIVE",
    "PAUSED",
    "ERROR",
    "REGION_CHOICES",
    "STATUS_CHOICES",
    "MoralisStreamsError",
    "MoralisStreamsCallFailed",
    "MoralisStreamsErrorReturned",
    "MoralisStreamsResponseFormatError",
    "MoralisStreamsApi",
]



# test cases for examples from https://github.com/MoralisWeb3/streams-beta#readme

import json
import logging
import time

import pytest
import requests

from moralis_streams_client import settings
from moralis_streams_client.signature import Signature
from moralis_streams_client.webhook import Webhook

WEBHOOK_TIMEOUT = 5


@pytest.fixture(scope="module", autouse=True)
def webhook_process(module_webhook_process):
    yield module_webhook_process


def test_webhook_hello(webhook, dump):
    ret = webhook.hello()
    assert ret == "Hello, World!"
    dump(ret)


def test_webhook_clear(webhook, dump):
    ret = webhook.clear()
    assert ret == "cleared"
    dump(ret)


def test_webhook_events(webhook, dump):
    ret = webhook.events()
    dump(ret)


def test_webhook_queue(webhook, dump, event_keys):
    webhook.clear()
    ret1 = webhook.inject(
        dict(
            int_data=1,
            str_data="some_string_1",
            dict_data={"foo": 1, "bar": 2},
        ),
    )
    assert ret1
    ret2 = webhook.inject(
        dict(
            int_data=2,
            str_data="some_string_2",
            dict_data={"foo": 1, "bar": 2},
        ),
    )
    assert ret2
    ret3 = webhook.inject(
        dict(
            int_data=2,
            str_data="some_string_3",
            dict_data={"foo": 1, "bar": 2},
        ),
    )
    assert ret3
    events = webhook.events()
    assert isinstance(events, list)
    for event in events:
        assert isinstance(event, dict)
        assert set(event.keys()) == set(event_keys)
        dump(event)


def test_webhook_tunnel(webhook, webhook_tunnel_url, dump):
    webhook.clear()
    url = webhook_tunnel_url + "/contract/event"
    logging.info(f"posting to url {url}")
    payload = {"message": "sent to public url"}
    headers = Signature().headers(payload)
    response = requests.post(url, json=payload, headers=headers)
    assert response.ok
    timeout = time.time() + WEBHOOK_TIMEOUT
    events = []
    while len(events) == 0:
        time.sleep(0.5)
        assert (
            time.time() < timeout
        ), "timeout waiting for webhook event callback"
        events = webhook.events()
        assert isinstance(events, list)
    assert len(events) == 1
    assert events[0]["body"] == payload
    dump(events)



from logging import info

import uvicorn

from . import settings
from .app import app
from .tunnel import NgrokTunnel


class ServerProcess:
    def __init__(self, **kwargs):
        self.addr = kwargs.get("addr", settings.ADDR)
        self.port = kwargs.get("port", settings.PORT)
        self.tunnel = kwargs.get("tunnel", settings.TUNNEL)
        self.token = kwargs.get("token", str(settings.NGROK_AUTHTOKEN))
        self.debug = kwargs.get("debug", settings.DEBUG)
        self.log_level = kwargs.get("log_level", settings.LOG_LEVEL)
        self.config = kwargs.get("config", {})
        self.config["tunnel"] = self.tunnel
        self.config["addr"] = self.addr
        self.config["port"] = self.port
        app.state.config = self.config
        app.state.tunnel_url = None

    def run(self):
        def fileConfig(*args, **kwargs):
            info("uvicorn logging config disabled")

        uvicorn.config.logging.config.fileConfig = fileConfig

        def _uvicorn_run():
            return uvicorn.run(
                "moralis_streams_client.app:app",
                host=self.addr,
                port=self.port,
                log_level=self.log_level.lower(),
                log_config="disable",
            )

        if self.tunnel is True:
            print(f"starting ngrok tunnel {self.tunnel=}")
            with NgrokTunnel(self.port, self.token) as tunnel:
                app.state.tunnel_url = tunnel.public_url
                return _uvicorn_run()
        else:
            print(f"ngrok tunnel disabled: {self.tunnel=}")
            return _uvicorn_run()
